# Add custom error messages related to DataLake
UNAUTHORIZED_DATALAKE_REQUEST_ERROR = (
    "A valid license is required to query Morningstar datalake. Please fill out this form for further assistance: https://go.morningstar.com/LP=6808"
)
UNAVAILABLE_EXTERNALLY = "Morningstar datalake is only accessible through Analytics Lab"
DATA_NOT_FOUND = "The requested data could not be found. Check the query parameters and try again."
TEMP_TABLE_NAME_NOT_FOUND = "One or more temporary tables does not have a name. Check the query parameters and try again."
